/*************************************
| rmv2rawvf version 0.1              |
| produces rawvf revision 5 from avf |
*************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#define VERSION 0.1

#define MAXREP 100000
#define MAXNAME 1000

struct event
{
	int time;
	int x,y;
	unsigned char event;
};
typedef struct event event;

FILE* RMV;

int mode,level,w,h,m;
int size;
int* board;
int* board_state;
int qm,nf;
int pause_allowed;
char name[MAXNAME];
char nick[MAXNAME];
char country[MAXNAME];
char skin[MAXNAME];
char program[MAXNAME];
char value[MAXNAME];
char timestamp[MAXNAME];
event video[MAXREP];
const int square_size=16;

void pause()
{
	if(!pause_allowed) return;
	fprintf(stderr,"Press enter to exit\n");
	while(getchar()!='\n');
}

void error(const char* msg)
{
	fprintf(stderr,"%s\n",msg);
	pause();
	exit(1);
}

char _fgetc(FILE* f)
{
	if(!feof(f)) return fgetc(f); else
	{
		error("Unexpected end of file");
	}
}

int getint(FILE* f)
{
	unsigned char c[4];
	int i;
	for(i=0;i<4;++i) c[i]=_fgetc(f);
	return (int)c[2]+c[3]*256+c[0]*65536+c[1]*16777216;
}

int getint2(FILE* f)
{
	unsigned char c[2];
	c[0]=_fgetc(f);c[1]=_fgetc(f);
	return (int)c[1]+c[0]*256;
}

int getint3(FILE* f)
{
	unsigned char c[3];
	c[0]=_fgetc(f);c[1]=_fgetc(f);c[2]=_fgetc(f);
	return (int)c[2]+c[1]*256+c[0]*65536;
}

void print_event(event* e)
{
	const char* event_names[]={"","mv","lc","lr","rc","rr","mc","mr","","pressed","pressedqm","closed",
		"questionmark","flag","blast","boom","won","nonstandard","number0","number1","number2","number3",
		"number4","number5","number6","number7","number8","blast"};
	unsigned char c=e->event;
	if(c<=7) // mouse event
	{
		printf("%d.%03d %s %d %d (%d %d)\n",
			e->time/1000,e->time%1000,
			event_names[c],
			e->x/square_size+1,e->y/square_size+1,
			e->x,e->y);
	}
	else if(c<=14 || (c>=18 && c<=27))	// board event
	{
		printf("%s %d %d\n",
			event_names[c],
			e->x,e->y);
	}
	else if(c<=17) // end event
	{
		printf("%s\n",
			event_names[c]);
	}
}
int readrmv()
{
	int i,j,cur=0;
	unsigned char c,d;
	const char* header_1="*rmv";
	const char* header_2="*umf";
	int fs;
	int result_string_size;
	int version_info_size;
	int player_info_size;
	int board_size;
	int preflags_size;
	int properties_size;
	int vid_size;
	int cs_size;
	int num_player_info;
	int name_length;
	int num_preflags;
	int is_first_event=1;
	
	// header 1
	for(i=0;i<4;++i) if((c=_fgetc(RMV))!=header_1[i] && c!=header_2[i]) error("No RMV header");
	if(getint2(RMV)!=1) error("Invalid video type");	
	fs=getint(RMV);
	
	// header 2
	result_string_size=getint2(RMV);
	version_info_size=getint2(RMV);
	player_info_size=getint2(RMV);
	board_size=getint2(RMV);
	preflags_size=getint2(RMV);
	properties_size=getint2(RMV);
	vid_size=getint(RMV);
	cs_size=getint2(RMV);
	_fgetc(RMV);	// newline
	
	// result string
	for(i=0;i<result_string_size-1;++i) _fgetc(RMV);
	// version info
	for(i=0;i<version_info_size-1;++i) program[i]=_fgetc(RMV);
	program[i]=0;
	_fgetc(RMV);	// a weird character 0x2E
	// player info
	num_player_info=getint2(RMV);
	if(num_player_info>0)
	{
		name_length=_fgetc(RMV);
		for(i=0;i<name_length;++i) name[i]=_fgetc(RMV);
		name[i]=0;
	}
	for(i=1;i<num_player_info;++i)	// i hope i got this correctly; all videos i have don't have this part
	{
		name_length=_fgetc(RMV);
		for(j=0;j<name_length;++j) _fgetc(RMV);
	}
	// board
	getint(RMV);
	w=_fgetc(RMV);
	h=_fgetc(RMV);
	m=getint2(RMV);
	board=(int*)malloc(sizeof(int)*w*h);
	for(i=0;i<w*h;++i) board[i]=0;
	for(i=0;i<m;++i)
	{
		c=_fgetc(RMV);d=_fgetc(RMV);
		if(c>w || d>h) error("Invalid mine position");
		board[d*w+c]=1;
	}
	// preflags
	if(preflags_size)
	{
		num_preflags=getint2(RMV);
		for(i=0;i<num_preflags;++i)
		{
			c=_fgetc(RMV);d=_fgetc(RMV);
			
			video[cur].event=4;
			video[cur].x=square_size/2+c*square_size;
			video[cur].y=square_size/2+d*square_size;
			video[cur++].time=0;
						
			video[cur].event=5;
			video[cur].x=square_size/2+c*square_size;
			video[cur].y=square_size/2+d*square_size;
			video[cur++].time=0;		
		}
	}
	// properties
	qm=_fgetc(RMV);
	nf=_fgetc(RMV);
	mode=_fgetc(RMV);
	level=_fgetc(RMV);
	for(i=4;i<properties_size;++i) _fgetc(RMV);
	
	// events: gogogo
	while(1)
	{
		video[cur].event=c=_fgetc(RMV);++i;
		if(!c) // timestamp event
		{
			getint(RMV);i+=4;
		}
		else if(c<=7) // mouse event
		{
			i+=8;
			video[cur].time=getint3(RMV);
			_fgetc(RMV);
			video[cur].x=getint2(RMV)-12;
			video[cur++].y=getint2(RMV)-56;
			
			if(is_first_event)
			{
				is_first_event=0;
				video[cur].event=video[cur-1].event;
				video[cur-1].event=2;
				video[cur].time=video[cur-1].time;
				video[cur].x=video[cur-1].x;
				video[cur++].y=video[cur-1].y;
			}
		}
		else if(c==8) error("Invalid event");
		else if(c<=14 || (c>=18 && c<=27))	// board event
		{
			i+=2;
			video[cur].x=_fgetc(RMV)+1;
			video[cur++].y=_fgetc(RMV)+1;
		}
		else if(c<=17) // end event
		{
			break;
		}
		else 
		{
			error("Invalid event");
		}
	}
	size=cur+1;

	return 1;
}

void writetxt()
{
	int i,j;
	const char* level_names[]={"beginner","intermediate","expert","custom"};
	const char* mode_names[]={"classic","UPK","cheat","density"};
	
	printf("RawVF_Version: Rev5\n");
	printf("Program: %s\n",program);
	printf("Player: %s\n",name);
	printf("Level: %s\n",level_names[level]);
	printf("Width: %d\n",w);
	printf("Height: %d\n",h);
	printf("Mines: %d\n",m);
	printf("Mode: %s\n",mode_names[mode]);
	if(qm) printf("Marks: on\n");
	printf("Board:\n");
	for(i=0;i<h;++i)
	{
		for(j=0;j<w;++j)
			if(board[i*w+j])
				printf("*");
			else
				printf("0");
		printf("\n");
	}
	printf("Events:\n");
	printf("0.000 start\n");
	for(i=0;i<size;++i)
	{
		print_event(video+i);
	}
	
}

int main(int argc,char** argv)
{
	if(argc<2)
	{
		printf("Usage: %s <input rmv> [nopause]\n",argv[0]);
		pause();
		return 0;
	}
	
	pause_allowed=argc==2;
	RMV=fopen(argv[1],"rb");
	
	if(!RMV)
	{
		printf("Couldn't open rmv\n");
		return 1;
	}
	
	if(!readrmv())
	{
		printf("Invalid RMV\n");
		return 1;
	}
	writetxt();
	fclose(RMV);free(board);
	
	pause();
	return 0;
}

